(function() {

    var AudioLoader = function() {

        var self = this;
        var sources = [];

        self.addSource = function(src) {
            sources.push(src);
            return self;
        };

        self.addSources = function(srcs) {
            srcs.forEach(function(src) {
                self.addSource(src);
            });
            return self;
        }

        self.load = function(callback) {
            var audios = {};
            var toLoad = sources.length;
            var loaded = 0;

            if (toLoad == 0) { callback(audios); return; }

            if (Game.Support.Embeded) {
                var isAndroid = navigator.userAgent.toLocaleLowerCase().indexOf("android") != -1;
                var prePath = isAndroid ? "/android_asset/www/" : "";
                sources.forEach(function(source) {
                    var media = new Media(prePath + (source.src || source));
                    setTimeout(function() {
                        audios[(source.name || source.src || source)] = media;
                        loaded++;
                        if (loaded == toLoad) {
                            callback(audios);
                        }
                    }, 150);
                });
            } else {
                callback(null);
            }
        };

    };

    window.Game.AudioLoader = AudioLoader;

    var AudioSFX = function(audio) {
        var self = this;
        self.play = function() {
            if (!Game.Audio._enabledSFX || !audio) return;
            try {
                self.setVolume(Game.Audio._volumeSFX);
                audio.play();
            } catch (ex) {
                console.log(ex);
            }
        };
        self.stop = function() {
            if (!audio) return;
            audio.stop();
        };
        self.setVolume = function(volume) {
            if (!audio) return;
            audio.setVolume(Math.clamp(volume, 0, 1));
            //audio.volume = Math.clamp(volume, 0, 1);
        };
    };

    window.Game.AudioSFX = AudioSFX;

    var AudioMusic = function(audio) {
        var self = this;

        self.play = function() {
            if (!Game.Audio._enabledMusic || !audio) return;
            try {
                audio.successCallback = null;
                audio.stop();
                audio.play();
                self.setVolume(Game.Audio._volumeMusic);
            } catch (ex) {
                console.log(ex);
            }
        };
        self.loop = function() {
            if (!Game.Audio._enabledMusic || !audio) return;
            try {
                //audio.stop();
                audio.successCallback = function() {
                    audio.play();
                };
                audio.play();
                setTimeout(function() {
                    self.setVolume(Game.Audio._volumeMusic);
                }, 100);
            } catch (ex) {
                console.log(ex);
            }
        };
        self.stop = function() {
            if (!audio) return;
            audio.successCallback = null;
            audio.stop();
        };
        self.setVolume = function(volume) {
            if (!audio) return;
            audio.setVolume(Math.clamp(parseFloat(volume), 0.01, 1));
        };
    };

    window.Game.AudioMusic = AudioMusic;



})();