(function() {

    var gameNode = Game.Data._rootNode;
    var selectedPiece = null;

    var startX = 0,
        startY = 0;

    var lastX = 0,
        lastY = 0;

    var deltaX = 0,
        deltaY = 0;

    var wasMoved = false;

    //

    var nextSnapPoint = null;
    var selectedSnapPoint = null;
    var moveVertical = false,
        movePositive = false;

    //

    var Controllers = {

        Puzzle: {

            OnInteractionStart: function(event) {
                // Only handle this when there is no selectedPiece.
                // No multi-touch for now.
                if (selectedPiece != null) {
                    Debug.Log("There is already a Piece selected. Only one Piece at a time, for now!");
                    return;
                }

                // Get the touch / start location
                var point = getXYFromEvent(event);

                // z-ordering pieces into a temp list
                var zOrderedPieces = Game.Data._pieces.copy();
                zOrderedPieces.sort(function(a, b) {
                    return b.z - a.z;
                });

                // checking if a piece can handle it
                // top-most z gets first chance
                for (var i = 0; i < zOrderedPieces.length; i++) {
                    if (zOrderedPieces[i].canBePickedAt(point.x, point.y)) {
                        selectedPiece = zOrderedPieces[i];
                        break;
                    }
                }

                // Nothing found? No further work to do
                if (!selectedPiece) {
                    return;
                }

                event.preventDefault();

                // pick up logic
                var startPos = selectedPiece.getPosition();
                startX = startPos.x;
                startY = startPos.y;
                lastX = point.x;
                lastY = point.y;
                deltaX = 0;
                deltaY = 0;
                wasMoved = false;

                Game.Audio.SFXPickup.play();

                sendPieceToForeground(selectedPiece);
                selectedPiece.pickup();
            },// OnInteractionStart

            OnInteractionMove: function(event) {

                // No Piece, no movement needed
                if (!selectedPiece) return;

                event.preventDefault();

                var point = getXYFromEvent(event);

                deltaX += (point.x - lastX);
                deltaY += (point.y - lastY);

                if (!wasMoved) {
                    if (Math.abs(deltaX) > 1 || Math.abs(deltaY) > 1) {
                        wasMoved = true;
                    }
                }

                lastX = point.x;
                lastY = point.y;

                var nextPointX = startX + deltaX,
                    nextPointY = startY + deltaY;

                // check x bounds
                if (nextPointX < -selectedPiece.width * 0.5) {
                    nextPointX = -selectedPiece.width * 0.5;
                } else if (nextPointX > Game.Data.GameAreaSize.width - selectedPiece.width * 0.5) {
                    nextPointX = Game.Data.GameAreaSize.width - selectedPiece.width * 0.5;
                }

                // check y bounds
                if (nextPointY < selectedPiece.offset) {
                    nextPointY = selectedPiece.offset;
                } else if (nextPointY > Game.Data.GameAreaSize.height) {
                    nextPointY = Game.Data.GameAreaSize.height;
                }

                selectedPiece.setPosition(nextPointX, nextPointY);
            },// OnInteractionMove

            OnInteractionEnd: function(event) {
                // No Piece, no drop
                if (!selectedPiece) return;

                event.preventDefault();

                var snapped = snap(selectedPiece);
                var hasSnapped = snapped != null;
                var shouldRotate = Game.Data.Rotatable && !wasMoved;

                selectedPiece.drop(snapped);

                if (hasSnapped) {
                    Game.Audio.SFXSnap.play();
                    if (!shouldRotate) {
                        sendPieceToBackground(selectedPiece);
                    }
                    setTimeout(function() {
                        checkWinCondition();
                    }, 20);
                }

                if (shouldRotate) {
                    (function(piece, isSnapped) {
                        piece.rotate(function() {
                            if (isSnapped) {
                                sendPieceToBackground(piece);
                            }
                        });
                    })(selectedPiece, hasSnapped);
                    if (!hasSnapped) Game.Audio.SFXRotate.play();
                } else {
                    if (!hasSnapped) Game.Audio.SFXDrop.play();
                }

                selectedPiece = null;
            }// OnInteractionEnd

        },

        SlidingPuzzle: {

            OnInteractionStart: function(event) {
                if (selectedPiece != null) {
                    Debug.Log("There is already a Piece selected. Only one Piece at a time, for now!");
                    this.OnInteractionEnd(event);
                }

                nextSnapPoint = null;
                selectedSnapPoint = null;

                var point = getXYFromEvent(event);

                for (var i = 0; i < Game.Data._pieces.length; i++) {
                    if (Game.Data._pieces[i].canBePickedAt(point.x, point.y)) {
                        selectedPiece = Game.Data._pieces[i];
                        break;
                    }
                }

                if (!selectedPiece) {
                    return;
                }

                if (!selectedPiece.snapPoint) {
                    throw new Error("Piece has no snap point!");
                }

                for (var dx = -1; (dx < 2 && nextSnapPoint == null); dx++) {
                    if (dx == 0) continue;
                    var nextGridX = selectedPiece.snapPoint._gridX + dx;
                    var sp = Game.Data.GetSnapPointAt(nextGridX, selectedPiece.snapPoint._gridY);
                    if (sp != null && sp.piece == null) {
                        nextSnapPoint = sp;
                    }
                }

                for (var dy = -1; (dy < 2 && nextSnapPoint == null); dy++) {
                    if (dy == 0) continue;
                    var nextGridY = selectedPiece.snapPoint._gridY + dy;
                    var sp = Game.Data.GetSnapPointAt(selectedPiece.snapPoint._gridX, nextGridY);
                    if (sp != null && sp.piece == null) {
                        nextSnapPoint = sp;
                    }
                }

                if (!nextSnapPoint) {
                    return;
                }

                event.preventDefault();

                // pick up logic
                var startPos = selectedPiece.getPosition();
                startX = startPos.x;
                startY = startPos.y;
                lastX = point.x;
                lastY = point.y;
                deltaX = 0;
                deltaY = 0;
                wasMoved = false;

                Game.Audio.SFXPickup.play();

                sendPieceToForeground(selectedPiece);
                selectedSnapPoint = selectedPiece.snapPoint;
                selectedPiece.tap();

                moveVertical = nextSnapPoint._gridX == selectedSnapPoint._gridX;
                movePositive = moveVertical ? (nextSnapPoint._gridY > selectedSnapPoint._gridY)
                    : (nextSnapPoint._gridX > selectedSnapPoint._gridX);

            },// OnInteractionStart

            OnInteractionMove: function(event) {
                // No Piece, no movement needed
                if (!selectedPiece || !nextSnapPoint) return;

                event.preventDefault();

                var point = getXYFromEvent(event);

                deltaX += (point.x - lastX);
                deltaY += (point.y - lastY);

                if (!wasMoved) {
                    if (Math.abs(deltaX) > 1 || Math.abs(deltaY) > 1) {
                        wasMoved = true;
                    }
                }

                lastX = point.x;
                lastY = point.y;

                var nextPointX = startX + deltaX,
                    nextPointY = startY + deltaY;

                // check x bounds
                if (nextPointX < -selectedPiece.width * 0.5) {
                    nextPointX = -selectedPiece.width * 0.5;
                } else if (nextPointX > Game.Data.GameAreaSize.width - selectedPiece.width * 0.5) {
                    nextPointX = Game.Data.GameAreaSize.width - selectedPiece.width * 0.5;
                }

                // check y bounds
                if (nextPointY < selectedPiece.offset) {
                    nextPointY = selectedPiece.offset;
                } else if (nextPointY > Game.Data.GameAreaSize.height) {
                    nextPointY = Game.Data.GameAreaSize.height;
                }





                if (nextSnapPoint._gridY == selectedSnapPoint._gridY) {
                    // next snap point is on horizontal plane, can't change y coord.
                    nextPointY = selectedPiece.y;
                    nextPointX = nextPointX.clamp(selectedSnapPoint.x, nextSnapPoint.x);
                }

                if (nextSnapPoint._gridX == selectedSnapPoint._gridX) {
                    // next snap point is on vertical plane, can't change x coord.
                    nextPointX = selectedPiece.x;
                    nextPointY = nextPointY.clamp(selectedSnapPoint.y, nextSnapPoint.y);
                }

                selectedPiece.setPosition(nextPointX, nextPointY);
            },// OnInteractionMove

            OnInteractionEnd: function(event) {
                // No Piece, no drop
                if (!selectedPiece) return;

                event.preventDefault();

                var snapped = snapToNearest(selectedPiece, selectedSnapPoint, nextSnapPoint);
                var hasSnapped = snapped != null;

                selectedPiece.drop(snapped);

                if (hasSnapped) {
                    Game.Audio.SFXSnap.play();
                    sendPieceToBackground(selectedPiece);
                    setTimeout(function() {
                        checkWinCondition();
                    }, 20);
                }

                if (!hasSnapped) {
                    Game.Audio.SFXDrop.play();
                }

                selectedPiece = null;
                nextSnapPoint = null;
                selectedSnapPoint = null;
            }// OnInteractionEnd

        }

    };// Controllers

    var CurrentController = Controllers.Puzzle;

    function GameNodeOnInteractionStart(event) {
        CurrentController.OnInteractionStart(event);
    }

    function GameNodeOnInteractionMove(event) {
        CurrentController.OnInteractionMove(event);
    }

    function GameNodeOnInteractionEnd(event) {
        CurrentController.OnInteractionEnd(event);
    }

    function snap(piece) {
        for (var i = 0; i < Game.Data.SnapPoints.length; i++) {
            var spot = Game.Data.SnapPoints[i];
            if (spot.piece != null) {
                continue;
            }
            var distanceSQ = spot.distanceSQ(piece.x, piece.y);
            if (distanceSQ <= 1000) {
                return spot;
            }
        }
        return null;
    }

    function snapToNearest(piece, snapA, snapB) {
        if (snapA == null || snapB == null) {
            return snapA != null ? snapA : snapB;
        }
        var distanceASQ = snapA.distanceSQ(piece.x, piece.y),
            distanceBSQ = snapB.distanceSQ(piece.x, piece.y);
        return (distanceASQ < distanceBSQ) ? snapA : snapB;
    }


    function checkWinCondition() {
        var inPlace = true;

        for (var i = 0; i < Game.Data._pieces.length && inPlace; i++) {
            inPlace = Game.Data._pieces[i].isAtTargetPosition();
        }

        if (inPlace) {
            // won
            var ps = new Game.ParticleSystem(document.body);
            ps.start();

            try {
                Game.Audio.SFXWin.play();
            } catch (ex) {
                console.log(ex);
            }
        }
    }

    function sendPieceToForeground(piece) {
        Game.Data._pieces.forEach(function (otherPiece, index) {
            if (otherPiece.z < piece.z || piece.equals(otherPiece)) {
                return;
            }
            otherPiece.setZ(otherPiece.z - 1);
        });
        piece.setZ(Game.Data.MaxZIndex);
    }

    function sendPieceToBackground(piece) {
        Game.Data._pieces.forEach(function (otherPiece, index) {
            if (otherPiece.z >= piece.z || piece.equals(otherPiece)) {
                return;
            }
            otherPiece.setZ(otherPiece.z + 1);
        });
        piece.setZ(0);
    }


    if (Game.Support.Touch) {
        gameNode.addEventListener("touchstart", GameNodeOnInteractionStart, false);
        gameNode.addEventListener("touchmove", GameNodeOnInteractionMove, false);
        gameNode.addEventListener("touchend", GameNodeOnInteractionEnd, false);
        gameNode.addEventListener("touchcancel", GameNodeOnInteractionEnd, false);
        gameNode.addEventListener("touchleave", GameNodeOnInteractionEnd, false);
    } else {
        gameNode.addEventListener("mousedown", GameNodeOnInteractionStart, false);
        gameNode.addEventListener("mousemove", GameNodeOnInteractionMove, false);
        gameNode.addEventListener("mouseup", GameNodeOnInteractionEnd, false);
        gameNode.addEventListener("mouseleave", GameNodeOnInteractionEnd, false);
    }

    function getXYFromEvent(event) {
        var p = { x: 0, y: 0 };

        if (event.touches && event.touches.length > 0) {
            p.x = event.touches[0].pageX;
            p.y = event.touches[0].pageY;

        } else {
            p.x = event.pageX;
            p.y = event.pageY;
        }

        return p;
    }



    window.Game.GameController = {
        setMode: function(mode) {
            switch (mode) {
                case Game.Modes.SlidingPuzzle:
                    CurrentController = Controllers.SlidingPuzzle;
                    break;
                case Game.Modes.Puzzle:
                    CurrentController = Controllers.Puzzle;
                    break;
                default:
                    throw new Error("Mode '" + mode + "' is not supported!");
            }

        }
    };

})();