(function() {

    var Localization = function() {

        var locale_default = "default";     // The default locale key, e.g. "default"
        var locale = locale_default;        // The current locale key, e.g. "en", "de", ...

        var data = {};                      // The whole dictionary with all data
        var localData = {};                 // Only the dictionary for the current locale

        var self = this;


        /**
         * Initialize this Localization.
         *
         * Sets the locale to Browser's locale or "en" as a fallback.
         * @param {Object} jsonData
         */
        self.init = function(jsonData) {
            locale = self.userLocale();
            data = jsonData;
            localData = jsonData[locale];
        };


        /**
         * Returns the localized value for the given key. When there is
         * no localized value the default localized value will be returned.
         * @param {string} key
         * @returns {string} localized value
         */
        self.get = function(key) {
            var val = null;
            try { val = localData[key]; } catch (ex) {}
            if (val != null) { return val; }
            val = data[locale_default][key];
            return val;
        };


        /**
         * Returns the user's locale or "en" as a fallback.
         * @returns {string} locale
         */
        self.userLocale = function() {
            var l = (navigator.language || navigator.userLanguage || "en").toLocaleLowerCase();
            if (l.length > 2) {
                l = l.substr(0, 2);
            }
            return l;
        };

    };// Localization


    window.Localization = new Localization();

})();