(function() {

    var Particle = function(opts, container) {

        var self = this;
        var lifetime = 2;
        var currenttime = 0;
        var opacity = 1;
        var speed = Math.random() * 300 + 150;
        var pos = Game.Utils.GameAreaCenter();
        var rad = Math.random() * Math.PI * 2;
        var dir = {
            x: Math.cos(rad),
            y: Math.sin(rad)
        };

        pos.x -= 32;
        pos.y -= 32;

        self.node = document.createElement("div");
        self.node.setAttribute("class", "particle");
        container.appendChild(self.node);

        self.update = function(delta) {
            if (self.isDone()) return;

            currenttime += delta;

            var tf = currenttime / lifetime;

            pos.x += dir.x * speed * delta;
            pos.y += dir.y * speed * delta;

            opacity = 1.0 - tf;

            //self.node.style.left = pos.x + "px";
            //self.node.style.top = pos.y + "px";
            self.node.style.webkitTransform = "translate3d("+pos.x+"px,"+pos.y+"px,0px)";
            self.node.style.opacity = opacity;
        };

        self.stop = function() {
            if (self.node) {
                self.node.parentNode.removeChild(self.node);
                self.node = null;
            }
        };

        self.isDone = function() {
            return currenttime >= lifetime;
        };

        self.update(0);
    };

    var ParticleSystem = function(container) {

        var self = this;
        var _running = false;
        var _done = false;
        var _particles = [];
        var lastupdate = Date.now();

        self.start = function() {
            if (_running) return;
            _running = true;

            for (var i = 0; i < 20; i++) {
                var p = new Particle(null, container);
                _particles.push(p);
            }

            lastupdate = Date.now();


            var render = function () {
                if (!_done) {
                    window.requestAnimFrame(render);
                }

                var now = Date.now();
                var delta = (now - lastupdate) * 0.001;
                lastupdate = now;

                var donecount = 0;

                for (var i = 0; i < _particles.length; i++) {
                    _particles[i].update(delta);
                    if (_particles[i].isDone()) {
                        _particles[i].stop();
                        donecount += 1;
                    }
                }

                if (donecount == _particles.length) {
                    _done = true;
                }
            }

            render();
        };



    };

    Game.ParticleSystem = ParticleSystem;

})();