(function() {

    var SlidingPuzzleGenerator = function(puzzle) {

        var self = this;
        var pieceCount = Game.Utils.TileCountForPuzzleAndDifficulty(puzzle);
        var missingPieceCoord = Game.Utils.MissingTileIndexForDifficulty(puzzle);
        var pieceWidth = 0,
            pieceHeight = 0;
        var totalPieceCount = -1;
        var totalPiecesLoaded = -1;
        var snapPoints = [];
        var backdrop = null;
        var backdropPadding = 5;
        var image = null;

        var _currentGameAreaCenter = null;
        var _gameDivNode = null;
        var _gameDivParentNode = null;

        totalPieceCount = pieceCount.total();

        /**
         *
         * @param {Number} [times]
         * @param {Number} [mode]
         */
        self.shuffle = function(times, mode) {
            if (times == undefined) {
                times = totalPieceCount * totalPieceCount;
            }
            times = Math.max(1, times);
            shuffleByChangePosition(times);
        };

        self.load = function(callback) {

            function onPuzzleReady(_image, backdropSize, countX, countY) {
                /*_gameDivNode = document.querySelector("#game");
                _gameDivParentNode = divGame.parentNode;
                _gameDivParentNode.removeChild(_gameDivNode);*/

                image = _image;

                pieceWidth = (image.width / Game.Utils.PixelRatio) / pieceCount.x;
                pieceHeight = (image.height / Game.Utils.PixelRatio) / pieceCount.y;
                backdropSize = Math.floor(backdropSize / Game.Utils.PixelRatio);

                var tmpCanvas = document.createElement('canvas');
                tmpCanvas.width = pieceWidth * Game.Utils.PixelRatio;
                tmpCanvas.height = pieceHeight * Game.Utils.PixelRatio;
                var tmpContext = tmpCanvas.getContext("2d");

                Game.Data.ClearPieces();
                Game.Data.SnapPoints.length = 0;
                Game.Data.Rotatable = parseInt(Game.Cache.getDef("rotatable", 0)) != 0;

                var centerScreen = Game.Utils.GameAreaCenter();
                _currentGameAreaCenter = centerScreen;

                var backdropX = centerScreen.x - backdropSize / 2,
                    backdropY = centerScreen.y - backdropSize / 2;


                // Called when backdrop is created
                function onBackdropCreated(backdropOptions) {
                    backdropPadding = backdropOptions.padding || 0;
                    totalPiecesLoaded = 0;

                    for (var x = 0; x < pieceCount.x; x++) {
                        for (var y = 0; y < pieceCount.y; y++) {

                            var srcx = x * (pieceWidth * Game.Utils.PixelRatio),
                                srcy = y * (pieceHeight * Game.Utils.PixelRatio),
                                srcw = pieceWidth * Game.Utils.PixelRatio,
                                srch = pieceHeight * Game.Utils.PixelRatio,
                                desx = 0,
                                desy = 0,
                                desw = pieceWidth * Game.Utils.PixelRatio,
                                desh = pieceHeight * Game.Utils.PixelRatio;

                            tmpContext.drawImage(
                                image,  // image
                                srcx,   // sx - from source image
                                srcy,   // sy - from source image
                                srcw,   // sw - from source image
                                srch,   // sh - from source image
                                desx,   // dx - on canvas
                                desy,   // dy - on canvas
                                desw,   // dw - on canvas
                                desh    // dh - on canvas
                            );

                            var data = tmpCanvas.toDataURL();
                            addPiece(x, y, pieceWidth, pieceHeight, data, callback);
                        }
                    }
                }

                if (backdrop && backdrop.parentNode) {
                    backdrop.parentNode.removeChild(backdrop);
                }

                backdrop = Game.Data.CreateBackdrop(
                    Game.Data._rootNode,
                    backdropX,
                    backdropY,
                    backdropSize,
                    backdropSize,
                    {
                        onLoaded: onBackdropCreated,
                        pieceWidth: pieceWidth,
                        pieceHeight: pieceHeight,
                        pieceCountX: countX,
                        pieceCountY: countY
                    }
                );


            }

            preparePuzzle(onPuzzleReady);
        };

        self.onExit = function() {

        };

        self.onResize = function(newSize, oldSize, callback) {
            var oldSnapPoints = [];
            Game.Data._pieces.forEach(function(piece) {
                oldSnapPoints.push(piece.snapPoint);
            });

            self.load({
                done: function() {
                    Game.Data.SnapPoints.forEach(function(snap) {
                        if (snap.piece != null) {
                            snap.piece.snapPoint = null;
                            snap.piece = null;
                        }
                    });
                    Game.Data._pieces.forEach(function(piece, index) {
                        if (piece.snapPoint) {
                            piece.snapPoint.piece = null;
                            piece.snapPoint = null;
                        }
                        var oldSnap = oldSnapPoints[index];
                        piece.snapPoint = Game.Data.GetSnapPointAt(oldSnap._gridX, oldSnap._gridY);
                        piece.snapPoint.piece = piece;
                        piece.setPosition(piece.snapPoint.x, piece.snapPoint.y);
                    });
                    if (callback) { callback(); }
                }
            });
        };

        self._onResize = function(newSize, oldSize) {
            /*var screenWidth = Game.Data._rootNode.clientWidth,
             screenHeight = Game.Data._rootNode.clientHeight;*/
            //var offset = self.calculateCenterOffset(Game.Data._rootNode);
            var offset = Game.Utils.GameAreaCenter();

            Game.Data.GameAreaSize = Game.Utils.GameAreaSize();

            var deltaSizeHalf = {
                width: (oldSize.width - newSize.width) * 0.5,
                height: (oldSize.height - newSize.height) * 0.5
            }

            //Game.Data.CreateBackdrop(Game.Data._rootNode, offset.x + 8, offset.y + 8, image.width, image.height);
            backdrop.style.left = (offset.x - image.width / 2 - 2) + "px";
            backdrop.style.top = (offset.y - image.height / 2 - 2) + "px";
            //Game.Data.CreateBackdrop(Game.Data._rootNode, offset.x - image.width * 0.5 - 2, offset.y - image.height * 0.5 - 2, image.width, image.height);
            Game.Data._pieces.forEach(function (piece, index) {
                var px = piece.x,
                    py = piece.y;

                px -= deltaSizeHalf.width;
                py -= deltaSizeHalf.height;

                if (px < 0) px = 0;
                if (px > newSize.width) px = newSize.width - piece.width;
                if (py < 0) py = 0;
                if (py > newSize.height) py = newSize.height - piece.height;

                piece.setPosition(px, py);
                piece.targetX = piece.targetX - deltaSizeHalf.width;
                piece.targetY = piece.targetY - deltaSizeHalf.height;
            });


        };

        function addPiece(x, y, w, h, src, callback) {
            var img = new Image();
            img.onload = function() {
                createPiece(x, y, w, h, img, callback);
            };
            img.src = src;
        }

        function createPiece(x, y, w, h, img, callback) {
            var shouldAdd = (missingPieceCoord.x != x || missingPieceCoord.y != y);
            var piece = null;

            if (shouldAdd) {
                piece = new Game.Piece({
                    parentNode: Game.Data._rootNode,
                    image: img,
                    x: 0,
                    y: 0,
                    width: img.width / Game.Utils.PixelRatio,
                    height: img.height / Game.Utils.PixelRatio,
                    zIndex: Game.Data._pieces.length
                });
            }

            var center = _currentGameAreaCenter;//Game.Utils.GameAreaCenter();
            var px = (x * w - Game.Piece.OFFSET_HALF / Game.Utils.PixelRatio) + center.x - image.width / Game.Utils.PixelRatio / 2;
            var py = (y * h - Game.Piece.OFFSET_HALF / Game.Utils.PixelRatio) + center.y - image.height / Game.Utils.PixelRatio / 2;

            px += backdropPadding * 0.5;
            py += backdropPadding * 0.5;

            if (shouldAdd) {
                piece.targetX = px;
                piece.targetY = py;
                piece.setPosition(px, py);
                piece.redraw();
                Game.Data.AddPiece(piece);
            }

            var snapPoint = {
                _gridX: x,
                _gridY: y,
                x: px,
                y: py,
                piece: shouldAdd ? piece : null,
                distanceSQ: function(x, y) {
                    var dx = x - this.x,
                        dy = y - this.y;
                    return (dx * dx) + (dy * dy);
                }
            };

            if (shouldAdd) {
                piece.snapPoint = snapPoint;
            }

            snapPoints.push(snapPoint);

            totalPiecesLoaded += 1;

            if (callback && callback.progress) {
                callback.progress(totalPiecesLoaded / totalPieceCount);
            }

            if (totalPiecesLoaded == totalPieceCount && callback && callback.done) {
                Game.Data.SnapPoints = snapPoints;
                Game.Data.GameAreaSize = Game.Utils.GameAreaSize();
                /*_gameDivParentNode.appendChild(_gameDivNode);
                _gameDivNode = null;
                _gameDivParentNode = null;*/
                callback.done();
            }
        }

        function shuffleByChangePosition(times) {
            for (var i = 0; i < times; i++) {
                slideToFreeSnapPoint();
            }

            var atLeastOneIsOffset = false;
            while (!atLeastOneIsOffset) {
                for (var i = 0; i < Game.Data._pieces.length; i++) {
                    var p = Game.Data._pieces[i];
                    atLeastOneIsOffset = !p.isAtTargetPosition();
                    if (atLeastOneIsOffset) {
                        break;
                    }
                }

                if (!atLeastOneIsOffset) {
                    slideToFreeSnapPoint();
                }
            }
        }

        function slideToFreeSnapPoint() {
            var freeSnap = Game.Data.GetFreeSnapPoint();
            var neighbors = Game.Data.GetNeighborSnapPoints(freeSnap);
            var selectedSnapPoint = neighbors.random();
            var piece = selectedSnapPoint.piece;
            selectedSnapPoint.piece = null;
            piece.snapPoint = freeSnap;
            freeSnap.piece = piece;
            piece.setPosition(freeSnap.x, freeSnap.y);
        }

        function preparePuzzle(callback) {
            var tileCount = Game.Utils.TileCountForPuzzleAndDifficulty(puzzle);
            var originalSize = -1;

            var img = new Image();
            img.onload = function() {

                if (originalSize == -1) { originalSize = Math.max(img.width, 1); }
                var scaleFactor = img.width / originalSize;

                var size = Game.Utils.GameAreaSize();
                var gameAreaSize = (Math.min(size.width, size.height) - 50) * Game.Utils.PixelRatio;

                var imageSize = Math.max(img.width, img.height);

                if (gameAreaSize < imageSize) {
                    var currentSize = gameAreaSize;
                    while ( (((currentSize / Game.Utils.PixelRatio / tileCount.x) % 2 != 0) || ((currentSize / Game.Utils.PixelRatio / tileCount.y) % 2 != 0)) && currentSize > 100 ) {
                        currentSize -= 1;
                    }
                    var c = document.createElement("canvas");
                    c.width = currentSize;
                    c.height = currentSize;
                    var g = c.getContext("2d");
                    g.drawImage(img, 0, 0, currentSize, currentSize);
                    img.src = c.toDataURL();
                    return;
                }

                callback(img,
                    Math.floor(1030 * scaleFactor),
                    tileCount.x,
                    tileCount.y);
            };
            img.src = "puzzles/" + puzzle.file;
        }
    };




    function randomInt(min, max) {
        return (Math.random() * (max - min) + min) | 0;
    }

    window.Game.SlidingPuzzleGenerator = SlidingPuzzleGenerator;
})();