if (!Array.prototype.copy) {
    Array.prototype.copy = function() {
        return this.slice();
    };
}

if (!Array.prototype.random) {
    Array.prototype.random = function() {
        if (this.length == 0) return null;
        return this[Math.floor(Math.random() * this.length)];
    };
}

if (!Array.prototype.removeAtIndex) {
    Array.prototype.removeAtIndex = function(index) {
        if (this.length == 0 || index >= this.length) return;
        var obj = this[index];
        this.splice(index, 1);
        return obj;
    };
}

if (!Math.clamp) {
    Math.clamp = function(num, min, max) {
        if (min > max) {
            var tmp = min;
            min = max;
            max = tmp;
        }
        return Math.min(Math.max(num, min), max);
    };
}

if (!Number.prototype.clamp) {
    Number.prototype.clamp = function(min, max) {
        if (min > max) {
            var tmp = min;
            min = max;
            max = tmp;
        }
        return Math.min(Math.max(this, min), max);
    };
}

(function() {
    window.requestAnimFrame = (function(){
        return  window.requestAnimationFrame       ||
            window.webkitRequestAnimationFrame ||
            window.mozRequestAnimationFrame    ||
            function( callback ){
                window.setTimeout(callback, 1000 / 60);
            };
    })();
})();


window.Debug = {
    Log: function(msg) {
        console.log(msg);
    }
};


window.Game = {
    AdMobIDAndroid: null,
    AdMobIDiOS: null,
    GetAdMobID: function() {
        if (Game.Support.Android) return this.AdMobIDAndroid;
        if (Game.Support.iOS) return this.AdMobIDiOS;
        return null;
    },
    Modes: {
        Puzzle: 'puzzle',
        SlidingPuzzle: 'sliding-puzzle'
    }
};

Game.Utils = {

    DevicePixelRatio: window['devicePixelRatio'] || 1,
    BackingStoreRatio: (function() {
        var c = document.createElement('canvas').getContext('2d');
        return c.webkitBackingStorePixelRatio ||
               c.mozBackingStorePixelRatio ||
               c.msBackingStorePixelRatio ||
               c.oBackingStorePixelRatio ||
               c.backingStorePixelRatio || 1
    })(),
    PixelRatio: 1,


    DrawImage: function(opts) {
        if(!opts.canvas) {
            throw("A canvas is required");
        }
        if(!opts.image) {
            throw("Image is required");
        }

        // get the canvas and context
        var canvas = opts.canvas,
            context = canvas.getContext('2d'),
            image = opts.image,

        // now default all the dimension info
            srcx = opts.srcx || 0,
            srcy = opts.srcy || 0,
            srcw = opts.srcw || image.naturalWidth,
            srch = opts.srch || image.naturalHeight,
            desx = opts.desx || srcx,
            desy = opts.desy || srcy,
            desw = opts.desw || srcw,
            desh = opts.desh || srch,
            auto = opts.auto,

        // finally query the various pixel ratios
            devicePixelRatio = window.devicePixelRatio || 1,
            backingStoreRatio = context.webkitBackingStorePixelRatio ||
                context.mozBackingStorePixelRatio ||
                context.msBackingStorePixelRatio ||
                context.oBackingStorePixelRatio ||
                context.backingStorePixelRatio || 1,

            ratio = devicePixelRatio / backingStoreRatio;

        // ensure we have a value set for auto.
        // If auto is set to false then we
        // will simply not upscale the canvas
        // and the default behaviour will be maintained
        if (typeof auto === 'undefined') {
            auto = true;
        }

        //context.save();

        // upscale the canvas if the two ratios don't match
        if (auto && devicePixelRatio !== backingStoreRatio) {
            if (!canvas.scaled) {
                var oldWidth = canvas.width;
                var oldHeight = canvas.height;

                canvas.width = oldWidth * ratio;
                canvas.height = oldHeight * ratio;

                canvas.style.width = oldWidth + 'px';
                canvas.style.height = oldHeight + 'px';

                canvas.scaled = true;
            }

            // now scale the context to counter
            // the fact that we've manually scaled
            // our canvas element
            context.scale(ratio, ratio);
        }

        context.drawImage(image, srcx, srcy, srcw, srch, desx, desy, desw, desh);

        //context.restore();
    },

    /**
     * @returns {Number}
     */
    GetTopBarHeight: function() {
        var v = parseInt(document.querySelector('#game div.topbar').offsetHeight);

        // When starting from home screen, game's topbar isn't visible, use home's instead
        if (v == 0) {
            v = parseInt(document.querySelector('#home div.topbar').offsetHeight);
        }

        return v;
    },

    /**
     * @returns {Number}
     */
    GetBottomBarHeight: function() {
        if (Game.Utils.ShowAd()) {
            return Math.round(100 / Game.Utils.PixelRatio);  // max height of Ad size
        }

        var v = parseInt(document.querySelector('#game div.bottombar').offsetHeight);

        // When starting from home screen, game's bottom isn't visible, use home's instead
        if (v == 0) {
            v = parseInt(document.querySelector('#home div.bottombar').offsetHeight);
        }

        return v;
    },

    GetScreenWidth: function() {
        return window.innerWidth;
    },

    GetScreenHeight: function() {
        return window.innerHeight;
    },

    ShowAd: function() {
        return Game.Ad && Game.AdMobIDAndroid && navigator.userAgent.toLocaleLowerCase().indexOf("android 4.1.1") == -1;
    },

    IsAdVisible: function() {
        if (Game.Ad) {
            return Game.Ad.visible;
        }
        return false;
    },

    GameAreaSize: function() {
        var size = { width: 0, height: 0 };
        size.width = this.GetScreenWidth();
        size.height = this.GetScreenHeight() - this.GetTopBarHeight() - this.GetBottomBarHeight();// - (Game.Utils.ShowAd() && Game.Utils.IsAdVisible() ? 0 : 100);
        return size;
    },

    GameAreaCenter: function() {
        var p = { x: 0, y: 0 };
        var size = this.GameAreaSize();
        p.x = size.width / 2;
        p.y = size.height / 2 + this.GetTopBarHeight();
        return p;
    },

    ForceDOMRedraw: function(element) {
        if (!element) return;

        var n = document.createTextNode(' ');
        var oldDisp = element.style.display;

        element.appendChild(n);
        element.style.display = "none";
        element.style.display = oldDisp;
        element.removeChild(n);
    },

    /**
     * @param [difficulty]
     * @returns {{x: number, y: number, total: Function}}
     * @constructor
     */
    TileCountForDifficulty: function(difficulty) {
        if (difficulty == null || difficulty == undefined) {
            difficulty = parseInt(Game.Cache.getDef("difficulty", 1));
        }

        var count = { x: 2, y: 2, total: function() { return this.x * this.y; }};
        switch (difficulty) {
            case 1:
                count.x = 3;
                count.y = 3;
                break;
            case 2:
                count.x = 5;
                count.y = 5;
                break;
            case 3:
                count.x = 10;
                count.y = 10;
                break;
        }
        return count;
    },

    TileCountForPuzzleAndDifficulty: function(puzzle, difficulty) {
        if (difficulty == null || difficulty == undefined) {
            difficulty = parseInt(Game.Cache.getDef("difficulty", 1));
        }

        if (Game.Data.Mode == Game.Modes.Puzzle) {
            return this.TileCountForDifficulty(difficulty);
        }

        if (puzzle['slidingTileSizes'] && difficulty < puzzle.slidingTileSizes.length) {
            var num = puzzle.slidingTileSizes[difficulty];
            return { x: num, y: num, total: function() { return this.x * this.y; }};
        }

        return this.TileCountForDifficulty(difficulty);
    },

    MissingTileIndexForDifficulty: function(puzzle, difficulty) {
        if (difficulty == null || difficulty == undefined) {
            difficulty = parseInt(Game.Cache.getDef("difficulty", 1));
        }

        var tileCount = this.TileCountForPuzzleAndDifficulty(puzzle, difficulty);
        var missingPos = {x: 0, y: 0};
        var index = 0;

        if (puzzle['slidingMissingIndices'] && difficulty < puzzle.slidingMissingIndices.length) {
            index = puzzle.slidingMissingIndices[difficulty] - 1;
        } else {
            index = Math.floor(Math.random() * (tileCount.x * tileCount.y));
        }

        missingPos.x = index % tileCount.x;
        missingPos.y = Math.floor(index / tileCount.x);

        return missingPos;
    }

};
Game.Utils.PixelRatio = Math.max(1, Math.floor(Game.Utils.DevicePixelRatio / Game.Utils.BackingStoreRatio));

Game.Data = {
    _pieces: [],
    _backdrop: null,
    _rootNode: document.getElementById("game"),
    GameAreaSize: null,
    Rotatable: false,
    Mode: null,
    SnapPoints: [], // {x, y, piece}
    PuzzleData: {
        current: -1
    },
    MaxZIndex: 0,

    /**
     * @param node
     * @param {number} x
     * @param {number} y
     * @param {number} w
     * @param {number} h
     * @param {Object} opts
     * @returns {HTMLCanvasElement} backdrop
     * @constructor
     */
    CreateBackdrop: function(node, x, y, w, h, opts) {
        if (this._backdrop) {
            this._backdrop.parentNode.removeChild(this._backdrop);
            this._backdrop = null;
        }

        if (!opts) opts = {};

        w = Math.floor(w);
        h = Math.floor(h);

        var bg = document.createElement("canvas");
        var bgX = x;
        var bgY = y;
        bg.style.position = "absolute";
        bg.style.width = (w + 0) + "px";
        bg.style.height = (h + 0) + "px";
        bg.width = w * Game.Utils.PixelRatio;
        bg.height = h * Game.Utils.PixelRatio;

        var img = new Image();
        img.onload = function() {
            var borderPadding = Math.round(5 * (w / img.width));

            bgX += borderPadding / 2;
            bgY += borderPadding / 2;

            bg.style.left = bgX + "px";
            bg.style.top = bgY + "px";

            var g = bg.getContext('2d');
            g.drawImage(img, 0, 0, img.width, img.height, 0, 0, bg.width, bg.height);



            if (opts.pieceWidth !== undefined) {
                var lx = borderPadding;
                var ly = borderPadding;
                var lw = 1;
                var lh = bg.height - (borderPadding * 2 * Game.Utils.PixelRatio);

                for (var i = 0; i < opts.pieceCountX - 1; i++) {
                    lx += (opts.pieceWidth * Game.Utils.PixelRatio);
                    g.fillStyle = "rgba(0, 0, 0, 0.75)";
                    g.fillRect(lx, ly, lw, lh);
                    g.fillStyle = "rgba(255, 255, 255, 0.15)";
                    g.fillRect(lx + lw, ly, lw, lh);
                }

                lx = borderPadding;
                ly = borderPadding;
                lw = bg.width - (borderPadding * 2 * Game.Utils.PixelRatio);
                lh = 1;
                for (var i = 0; i < opts.pieceCountY - 1; i++) {
                    ly += (opts.pieceHeight * Game.Utils.PixelRatio);
                    g.fillStyle = "rgba(0, 0, 0, 0.75)";
                    g.fillRect(lx, ly, lw, lh);
                    g.fillStyle = "rgba(255, 255, 255, 0.15)";
                    g.fillRect(lx, ly + lh, lw, lh);
                }

            }


            if (opts.onLoaded) {
                opts.onLoaded({
                    padding: borderPadding
                });
            }
        };
        img.src = 'puzzles/Spielflaeche.png';



        /*this._backdrop = document.createElement("div");
        this._backdrop.setAttribute("class", "backdrop");
        this._backdrop.style.width = (w + 0) + "px";
        this._backdrop.style.height = (h + 0) + "px";
        this._backdrop.style.top = (y + 0) + "px";
        this._backdrop.style.left = (x + 0) + "px";
        node.appendChild(this._backdrop);*/

        this._backdrop = bg;
        node.appendChild(this._backdrop);
        return this._backdrop;
    },
    AddPiece: function(piece) {
        this.MaxZIndex = this._pieces.length;
        this._pieces.push(piece);
    },
    ClearPieces: function() {
        if (this._backdrop) {
            this._backdrop.parentNode.removeChild(this._backdrop);
            this._backdrop = null;
        }
        this._pieces.forEach(function(piece) {
            piece.delete();
        });
        this._pieces.length = 0;
        this.MaxZIndex = 0;
    },

    /**
     * @return {object}
     */
    GetSnapPointAt: function(gridX, gridY) {
        if (gridX < 0 || gridY < 0) return null;
        for (var i = 0; i < Game.Data.SnapPoints.length; i++) {
            var sp = Game.Data.SnapPoints[i];
            if (sp._gridX == gridX && sp._gridY == gridY) {
                return sp;
            }
        }
        return null;
    },

    /**
     * @return {object}
     */
    GetFreeSnapPoint: function() {
        for (var i = 0; i < Game.Data.SnapPoints.length; i++) {
            if (Game.Data.SnapPoints[i].piece == null) {
                return Game.Data.SnapPoints[i];
            }
        }
        return null;
    },

    /**
     * @return {Array<object>}
     */
    GetNeighborSnapPoints: function(snapPoint) {
        var list = [];
        if (snapPoint) {
            var snap = Game.Data.GetSnapPointAt(snapPoint._gridX, snapPoint._gridY - 1); // top
            if (snap) list.push(snap);
            snap = Game.Data.GetSnapPointAt(snapPoint._gridX + 1, snapPoint._gridY);     // right
            if (snap) list.push(snap);
            snap = Game.Data.GetSnapPointAt(snapPoint._gridX, snapPoint._gridY + 1);     // bottom
            if (snap) list.push(snap);
            snap = Game.Data.GetSnapPointAt(snapPoint._gridX - 1, snapPoint._gridY);     // left
            if (snap) list.push(snap);
        }
        return list;
    }
};


Game.Support = {

    Touch: ('ontouchstart' in document.documentElement),
    Android: navigator.userAgent.toLocaleLowerCase().indexOf("android") != -1,
    iOS: !!(navigator.userAgent.match(/(iPad|iPhone|iPod)/g)),
    Embeded: (navigator.userAgent.toLocaleLowerCase().indexOf("android") != -1 || !!(navigator.userAgent.match(/(iPad|iPhone|iPod)/g))),
    /**
     * @returns {string}
     */
    ButtonActionEventName: function() {
        return this.Touch ? "touchend" : "click";
    }
};


Game.Ajax = {

    /* load */
    load: function(opts) {
        /*
        if (Game.Support.Embeded && opts.url.indexOf("http") != 0) {
            this._loadForCordova(opts);
            return;
        }*/
        this._loadForWeb(opts);
    },
    _loadForCordova: function(opts) {
        function formatError(error) {
            switch (error.code) {
                case FileError.NOT_FOUND_ERR: return "NOT_FOUND_ERR";
                case FileError.SECURITY_ERR: return "SECURITY_ERR";
                case FileError.ABORT_ERR: return "ABORT_ERR";
                case FileError.NOT_READABLE_ERR: return "NOT_READABLE_ERR";
                case FileError.ENCODING_ERR: return "ENCODING_ERR";
                case FileError.NO_MODIFICATION_ALLOWED_ERR: return "NO_MODIFICATION_ALLOWED_ERR";
                case FileError.INVALID_STATE_ERR: return "INVALID_STATE_ERR";
                case FileError.SYNTAX_ERR: return "SYNTAX_ERR";
                case FileError.INVALID_MODIFICATION_ERR: return "INVALID_MODIFICATION_ERR";
                case FileError.QUOTA_EXCEEDED_ERR: return "QUOTA_EXCEEDED_ERR";
                case FileError.TYPE_MISMATCH_ERR: return "TYPE_MISMATCH_ERR";
                case FileError.PATH_EXISTS_ERR: return "PATH_EXISTS_ERR";
            }
            return "Unkown Error: " + error.code;
        }

        function fail(error) {
            if (opts.failed) {
                opts.failed(formatError(error));
            }
        }

        window.requestFileSystem(
            LocalFileSystem.PERSISTENT,
            0,
            function(filesystem) {
                filesystem.root.getFile(
                    opts.url,
                    null,
                    function(fileentry) {
                        fileentry.file(function(file) {
                            var reader = new FileReader();
                            reader.onloadend = function(evt) {
                                opts.done(evt.target.result);
                            };
                            reader.readAsText(file);
                        },
                        fail);
                    },
                    fail);
            },
            fail);
    },
    _loadForWeb: function(opts) {


        if (Game.Support.Embeded && opts.url.indexOf("http") != 0 && opts.url.indexOf("file") != 0 && Game.Support.Android) {
            opts.url = "file:///android_asset/www/" + opts.url;
        }

        var req = new XMLHttpRequest();
        req.overrideMimeType((opts.mimeType || "application/json"));
        req.open((opts.method || "GET"), opts.url, true);
        req.onreadystatechange = function() {
            if (req.readyState == 4 && (req.status == 200 || req.status == 0)) {
                opts.done(req.responseText);
            }
        };
        var data = null;
        if (opts.data) data = opts.data;
        req.send(data);
    }
};