/*
TODO:
 - Hochauflösendes Display (puzzle offset wenn im portrait gestartet)
 - Puzzleteile teilweise nicht verschiebbar (nur in der Mitte anzufassen)

 - Versionsnummer richtig positionieren
 - Name wird bei sehr kleinen devices abgeschnitten

 - Reflex auf Spielsteine

DONE:
 - Size vom Device
 - Kacheln Schatten wenn kleiner abgeschnitten
 - Home-Screen Puzzleauswahl optimieren (Größe)
 */

(function () {

    var RUNTIME_VERSION = "6";

    Game.Audio = {
        _AllSFX: [],
        _AllMusic: [],
        _volumeMusic: parseFloat(Game.Cache.getDef("volumeMusic", 0.5)),
        _volumeSFX: parseFloat(Game.Cache.getDef("volumeSFX", 1.0)),
        _enabledMusic: Game.Cache.getDef("enabledMusic", "true") == "true",
        _enabledSFX: Game.Cache.getDef("enabledSFX", "true") == "true",
        setVolumeMusic: function (volume) {
            this._volumeMusic = volume;
            this._AllMusic.forEach(function (item) {
                item.setVolume(volume);
            });
        },
        setVolumeSFX: function (volume) {
            this._volumeSFX = volume;
            /*this._AllSFX.forEach(function (item) {
                item.setVolume(volume);
            });*/
        },
        toggleEnabledMusic: function () {
            var state = (this._enabledMusic = !this._enabledMusic);
            this._AllMusic.forEach(function (item) {
                if (state) {
                    item.loop();
                } else {
                    item.stop();
                }
            });
            return this._enabledMusic;
        },
        toggleEnabledSFX: function () {
            this._enabledSFX = !this._enabledSFX;
            return this._enabledSFX;
        },
        pause: function() {
            Game.Audio.MusicBackground.stop();
        },
        resume: function() {
            Game.Audio.MusicBackground.loop();
        }
    };


    // topbar        77px
    // bottombar    115px

    var imageLoader = new Game.ImageLoader();
    var Images = null;

    var audioLoader = new Game.AudioLoader();
    audioLoader.addSources([
        {"name": "SFXShake",        "src": "audio/shakey.wav"},
        {"name": "SFXPickup",       "src": "audio/pickup.wav"},
        {"name": "SFXDrop",         "src": "audio/putdown.wav"},
        {"name": "SFXRotate",       "src": "audio/rotate.wav"},
        {"name": "SFXTap",          "src": "audio/tap.wav"},
        {"name": "SFXSnap",         "src": "audio/lock.wav"},
        {"name": "SFXWin",          "src": "audio/win.mp3"},
        {"name": "MusicBackground", "src": "puzzles/Background.mp3"}
    ]);

    imageLoader.addSources([
        {src: "img/star.png", name: "star"}
    ]);

    Game.Ajax.load({
        url: "puzzles/puzzles.json",
        done: function(data) {

            try {

                var jsonData = JSON.parse(data);
                Game.AdMobIDAndroid = jsonData["AdMobIDAndroid"];
                Game.AdMobIDiOS = jsonData["AdMobIDiOS"];
                Game.Data.PuzzleData.puzzles = jsonData["Puzzles"];
                Game.Data.Mode = (jsonData["Mode"] || Game.Modes.Puzzle);
                Game.PuzzleChooserController.createButtons(Game.Data.PuzzleData.puzzles);

                // Deactivate rotation for sliding puzzles
                if (Game.Data.Mode == Game.Modes.SlidingPuzzle) {
                    document.querySelector("#settings tr.rotate td").style.display = "none";
                }

                if (jsonData["ShowLogo"] == "true") {
                    var logoList = document.querySelectorAll("div.topbar div.logo");
                    for (var i = 0; i < logoList.length; i++) {
                        logoList[i].style.display = "block";
                    }
                }
            } catch (ex) {
                console.log(ex);
            }

            imageLoader.load(function (loadedImages) {

                Images = loadedImages;
                Images.get = function (name) {
                    return this[name];
                };

                audioLoader.load(function (audios) {

                    if (audios == null) {
                        audios = {};
                    }
                    Game.Audio._AllMusic.push(Game.Audio.MusicBackground = new Game.AudioMusic(audios["MusicBackground"]));
                    Game.Audio._AllSFX.push(Game.Audio.SFXPickup = new Game.AudioSFX(audios["SFXPickup"]));
                    Game.Audio._AllSFX.push(Game.Audio.SFXDrop = new Game.AudioSFX(audios["SFXDrop"]));
                    Game.Audio._AllSFX.push(Game.Audio.SFXRotate = new Game.AudioSFX(audios["SFXRotate"]));
                    Game.Audio._AllSFX.push(Game.Audio.SFXTap = new Game.AudioSFX(audios["SFXTap"]));
                    Game.Audio._AllSFX.push(Game.Audio.SFXSnap = new Game.AudioSFX(audios["SFXSnap"]));
                    Game.Audio._AllSFX.push(Game.Audio.SFXWin = new Game.AudioSFX(audios["SFXWin"]));
                    Game.Audio._AllSFX.push(Game.Audio.SFXShake = new Game.AudioSFX(audios["SFXShake"]));

                    Game.Ajax.load({
                        url: 'puzzles/localization.json',
                        done: function(responseText) {
                            Localization.init(JSON.parse(responseText));
                            localizeMe();
                        }
                    });

                    function localizeMe() {
                        var eles = document.querySelectorAll("div.bottombar span");
                        var text = Localization.get("game_title");
                        for (var i = 0; i < eles.length; i++) {
                            eles[i].innerText = text;
                        }

                        // Translate titles
                        function translateTitle(selector, key, escapeHTML) {
                            escapeHTML = (escapeHTML == undefined) ? true : escapeHTML;

                            if (escapeHTML) {
                                document.querySelector(selector).innerText = Localization.get(key);
                            } else {
                                document.querySelector(selector).innerHTML = Localization.get(key);
                            }
                        }

                        // info
                        translateTitle("#info #imprint h3",          "imprint_title");
                        translateTitle("#info #imprint p.text1",    "imprint_text1", false);
                        translateTitle("#info #imprint p.text2",    "imprint_text2", false);
                        translateTitle("#info #imprint p.text3",    "imprint_text3", false);

                        // settings
                        /*translateTitle("#settings th.music",    "settings_Music");
                         translateTitle("#settings th.sound",    "settings_Sound");
                         translateTitle("#settings th.level",    "settings_Level");
                         translateTitle("#settings th.rotation", "settings_Rotation");*/

                        var loading = document.getElementById("loading");
                        //TweenMax.to(loading, 0.5, { delay: 0.5, opacity:0, onComplete: function() { loading.parentNode.removeChild(loading); } });
                        loading.parentNode.removeChild(loading);
                        loading = null;

                        Game.Audio.MusicBackground.loop();

                        (function() {
                            var versionNode = document.createElement("div");
                            versionNode.className = "app-version";
                            versionNode.innerText = Localization.get("app_version") + "_" + RUNTIME_VERSION;
                            document.querySelector("#settings div.topbar").appendChild(versionNode);
                        })();

                        Game.ShowAd();
                    }

                    //Game.Audio.MusicBackground.loop();
                    //Game.Start(0);


                });

            });
        }
    });



    var puzzleGenerator = null;

    Game._StartPuzzle = function(index, mode) {
        Game.HideAd(function() {
            Game.Data.PuzzleData.current = index;
            if (puzzleGenerator != null) {
                puzzleGenerator.onExit();
            }
            puzzleGenerator = new Game.PuzzleGenerator(Game.Data.PuzzleData.puzzles[index]);
            puzzleGenerator.load({
                done: function() {
                    puzzleGenerator.shuffle(5, mode);
                    Game.TopbarController.GameStarted();
                    setTimeout(function() {
                        Game.ShowAd();
                    }, 5000);
                }
            });
        });
        /*Game.Data.PuzzleData.current = index;

        var tileCount = Game.Utils.TileCountForDifficulty();
        var originalSize = -1;

        var img = new Image();
        img.onload = function() {
            if (originalSize == -1) { originalSize = Math.max(img.width, 1); }
            var scaleFactor = img.width / originalSize;

            var size = Game.Utils.GameAreaSize();
            var gameAreaSize = Math.min(size.width, size.height) - 50;

            var imageSize = Math.max(img.width, img.height);

            if (gameAreaSize < imageSize) {
                var currentSize = gameAreaSize;
                while ( (((currentSize / tileCount.x) % 2 != 0) || ((currentSize / tileCount.y) % 2 != 0)) && currentSize > 100 ) {
                    currentSize -= 1;
                }
                var c = document.createElement("canvas");
                c.width = currentSize;
                c.height = currentSize;
                var g = c.getContext("2d");
                g.drawImage(img, 0, 0, currentSize, currentSize);
                img.src = c.toDataURL();
                return;
            }

            if (puzzleGenerator != null) {
                puzzleGenerator.onExit();
            }

            puzzleGenerator = new Game.PuzzleGenerator(img);
            puzzleGenerator.load({
                done: function() {
                    puzzleGenerator.shuffle(5, mode);
                    Game.TopbarController.GameStarted();
                    Game.ShowAd();
                }
            }, Math.floor(1030 * scaleFactor), tileCount.x, tileCount.y);
        };
        img.src = "puzzles/" + Game.Data.PuzzleData.puzzles[index].file;*/
    };

    Game._StartSlidingPuzzle = function (index, mode) {
        Game.HideAd(function() {
            Game.Data.PuzzleData.current = index;
            if (puzzleGenerator != null) {
                puzzleGenerator.onExit();
            }
            puzzleGenerator = new Game.SlidingPuzzleGenerator(Game.Data.PuzzleData.puzzles[index]);
            puzzleGenerator.load({
                done: function() {
                    puzzleGenerator.shuffle();
                    Game.TopbarController.GameStarted();
                    setTimeout(function() {
                        Game.ShowAd();
                    }, 5000);
                }
            });
        });
    };

    Game.Start = function (index, mode) {
        Game.GameController.setMode(Game.Data.Mode);
        hideOrientationChangeOverlay();
        if (Game.Data.Mode == Game.Modes.Puzzle) {
            Game._StartPuzzle(index, mode);
            return;
        }

        if (Game.Data.Mode == Game.Modes.SlidingPuzzle) {
            Game._StartSlidingPuzzle(index, mode);
            return;
        }

        throw new Error("Unspecified game mode!");
    };

    Game.Next = function () {
        Game.HideAd(function() {
            var nextIndex = Game.Data.PuzzleData.current + 1;
            if (nextIndex >= Game.Data.PuzzleData.puzzles.length) {
                nextIndex = 0;
            }
            Game.Start(nextIndex);
        });
    };

    Game.Previous = function () {
        Game.HideAd(function() {
            var prevIndex = Game.Data.PuzzleData.current - 1;
            if (prevIndex < 0) {
                prevIndex = Game.Data.PuzzleData.puzzles.length - 1;
            }
            Game.Start(prevIndex);
        });
    };

    Game.Reset = function (mode) {
        Game.Start(Game.Data.PuzzleData.current, mode);
    };

    var orientationChangeOverlay = null;
    function showOrientationChangeOverlay() {
        if (orientationChangeOverlay != null) return;
        orientationChangeOverlay = document.createElement("div");
        orientationChangeOverlay.setAttribute("id", "orientation-overlay");
        document.body.appendChild(orientationChangeOverlay);
        TweenMax.fromTo(orientationChangeOverlay, 0.5, { alpha: 0 }, { alpha: 1 });
    }
    function hideOrientationChangeOverlay() {
        if (orientationChangeOverlay == null) return;

        function onComplete() {
            orientationChangeOverlay.parentNode.removeChild(orientationChangeOverlay);
            orientationChangeOverlay = null;
        }

        TweenMax.fromTo(orientationChangeOverlay, 0.5, { alpha: 1 }, { alpha: 0, onComplete: onComplete });
    }

    Game.Resize = function(newSize, oldSize) {
        if (puzzleGenerator == null) {
            Game.ShowAd();
            return;
        }

        puzzleGenerator.onResize(newSize, oldSize, function() {
            hideOrientationChangeOverlay();
            setTimeout(function() { Game.ShowAd(); }, 1000);
        });
    };

    Game.OnExit = function() {
        if (puzzleGenerator == null) {
            return;
        }

        puzzleGenerator.onExit();
        puzzleGenerator = null;
    };

    Game.ShowAd = function(callback) {
        if (!Game.Utils.ShowAd()) {
            if (callback) {
                callback();
            }
            return;
        }

        Game.Ad.loadAd(
            Game.GetAdMobID(),
            function(message) {
                if (message == Game.Ad.MESSAGE_AD_RECEIVED) {
                    document.body.className = "ad";
                    document.querySelector("#game div.bottombar").style.display = "none";

                    if (callback) {
                        callback();
                    }
                }
            },
            false
        );
    };

    Game.HideAd = function(callback) {
        if (!Game.Utils.ShowAd()) {
            if (callback) {
                callback();
            }
            return;
        }

        Game.Ad.hideAd(function() {
            document.body.className = "";
            document.querySelector("#game div.bottombar").style.display = "block";

            if (callback) {
                callback();
            }
        });
    };

    document.addEventListener("pause", function() {
        Game.Audio.pause();
    }, false);

    document.addEventListener("resume", function() {
        Game.Audio.resume();
    }, false);



// android:screenOrientation="landscape"

    var orientationChangeTimerID = null;

    function resizeAndOrientationChangeHandler(evt) {
        if (orientationChangeTimerID != null) {
            clearTimeout(orientationChangeTimerID);
        }

        if (orientationChangeTimerID == null) {
            Game.HideAd();

            if (puzzleGenerator != null) {
                showOrientationChangeOverlay();
            }
        }

        var oldSize = {
            width: Game.Data._rootNode.clientWidth,
            height: Game.Data._rootNode.clientHeight
        }

        orientationChangeTimerID = setTimeout(function() {
            var newSize = {
                width: Game.Data._rootNode.clientWidth,
                height: Game.Data._rootNode.clientHeight
            }
            Game.Resize(newSize, oldSize);
            orientationChangeTimerID = null;
        }, 1000);
    }

    if (Game.Support.Touch) {
        window.addEventListener("orientationchange", resizeAndOrientationChangeHandler, false);
    } else {
        window.addEventListener("resize", resizeAndOrientationChangeHandler, false);
    }

})();